unit MainForm;

{
Revision history:

1997 Jan 14  V0.0.0  First version
1997 Jan 16  V0.0.2  Allow timestamp instead of checksum
                     Use inherited file properties dialog box
1997 Jan 19  V1.0.0  Add icon display to file properties dialog box
                     Make timestamp the default method for speed and
                     integrity of file access date
1997 Jan 27  V1.0.2  Remove code timing the components of the compare phase
                     Store directory name separately from file name to save space
1997 Jan 29  V1.0.4  Clear directory list at end of compare phase
1997 Feb 03  V1.0.6  Make drives listbox owner-draw
                     Make floppy, if chosen, a "master" directory
1997 Feb 12  V1.0.8  Update FileProperties form, decode date/time 0 as "unknown"
1997 Apr 02  V1.1.0  Make file list box hint the filename (for long paths!)
                     Save and optionally restore duplicate file list
                     By default, ignore files in Win 95 SYSBCKUP directory
                     By default, ignore zero-length files
                     Replace ListBox with ListView (both Drives and Results)
                     Correct: missing FindClose in do_checksum routine
                     Correct: remove deleted file from the duplicates list
1997 Apr 07  V1.1.2  Use ShellAPI function to move file to recycle bin
1997 May 13  V1.1.4  Use my own TFileList component
                     Don't show properties/delete box for non-existant files
                     Put source files in sub-folder
                     Force checksum routine to return 31-bit value
1997 May 23  V1.2.0  Move to Delphi 3.0
                     Don't leave singletons in the duplicates list
                     Correct property display for sequential compressed files
                     Don't allow ColumnClick on the FileListView - set False
1997 Oct 08  V1.2.2  Move to Delphi 3.01
                     Handle large font displays better.
                     Use TreeScanner with FindHiddenXX options
                     Don't build against run-time VCL30.DPL
1997 Nov 24  V1.2.4  Move to Delphi 3.02
                     Do build against run-time VCL30.DPL
1997 Nov 29  V1.2.6  Allow a list of folders to ignore
1998 Jan 24  V1.2.8  Save and restore ignore folder list using registry
                     Save and restore edit and check box settings
                     Save a list of found duplicate file names as "NewDupl.lis"
                     Optionally read "SkipDupl.lis" an ignore list on next scan
                       (pending drag-and-drop generation of this list)
                     Skip special folders that have DESKTOP.INI present
                     Add "File names must be same" option (disabled by default)
1998 Feb 23  V1.3.0  Add file version column
1998 Feb 28  V1.3.2  Allow for scanning multiple root folders
                     Handle zero matching files found correctly (oops!)
                     Add scan details to program caption
1998 Mar 15  V1.4.0  Allow finding multiple file masks
                     Allow for ignoring file size (i.e show all DLL versions)
                     Add auto-size of displayed column widths
1998 Mar 16  V1.4.2  Correct problems in FileListView when columns missing
                     Sort "find-all" (ignore size) output into file name order
                     Add program version information
1998 May 07  V1.4.4  Make skipping of DESKTOP.INI folders optional
                     Split search controls into Basic and Advanced
                     Improve ListView drawing speed
                     Don't lower-case file names
                     Make skipping of DESKTOP.INI folders optional
                     Add View button, uses application or QuickView if installed
1998 Jun 19  V1.4.4  Simply update the ReadMe.txt to have a Web URL
1998 Jul 02  V1.4.6  Add delete & shift/delete key handling ...
                     ... based on code from Chris van den Hout
1998 Oct 09  V2.0.0  Version for Delphi 4.01
                     Remove multi-select flag from file ListView as we don't
                       allow multi-delete
                     Make list view columns zero width while data is read from file
1998 Oct 13  V2.0.2  Clear "UseSystemFont" from StatusBar
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, TreeScan, ComCtrls, ExtCtrls, DiskList, FileList, VersInfo;

type
  TChecksum = DWORD;   // for rapid checksum on the first 512 bytes of a file

type
  TfrmMain = class(TForm)
    StatusBar1: TStatusBar;
    GroupBox1: TGroupBox;
    edtFileMask: TEdit;
    Label2: TLabel;
    btnSearch: TButton;
    edtRootDir: TEdit;
    Label7: TLabel;
    Timer1: TTimer;
    btnExit: TButton;
    TreeScanner1: TTreeScanner;
    DiskListView1: TDiskListView;
    edtSkipFolder: TEdit;
    Label1: TLabel;
    lblFileListCaption: TLabel;
    VersionInfoResource1: TVersionInfoResource;
    lblVersion: TLabel;
    PageControl1: TPageControl;
    TabSheetBasic: TTabSheet;
    TabSheetAdvanced: TTabSheet;
    chkSkipSysbckup: TCheckBox;
    chkCheckFilesize: TCheckBox;
    chkFloppyIsMaster: TCheckBox;
    chkCheckFileTimestamp: TCheckBox;
    chkCheckFilename: TCheckBox;
    chkSkipDesktopFolders: TCheckBox;
    chkSkipEmptyFiles: TCheckBox;
    FileListView1: TFileListView;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure TreeScanner1FileFound(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure TreeScanner1DirectoryFound(Sender: TObject);
    procedure btnSearchClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure btnExitClick(Sender: TObject);
    procedure TreeScanner1DirectoryDone(Sender: TObject);
    procedure FileListView1DblClick(Sender: TObject);
    procedure FileListView1Click(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure DiskListView1Change(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure edtRootDirExit(Sender: TObject);
    procedure FileListView1ColumnClick(Sender: TObject;
      Column: TListColumn);
    procedure chkCheckFilesizeClick(Sender: TObject);
    procedure FileListView1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    file_list: TStringList;         // list of files, with pointer to folder
    dir_list: TStringList;          // list of folders
    master_list: TStringList;       // simple list of ASCII sizes
    skip_folder_list: TStringList;  // list of folders to skip
    skip_file_list: TStringList;
    current_dir: integer;           // index of current folder in its list
    comparing: boolean;             // true when a compare in progress
    continue_compare: boolean;      // user sets this false to stop a scan
    duplicate_bytes: integer;       // total number of bytes (excluding "master")
    duplicate_files: integer;       // total number of duplicate files
    saved_duplicates_filename: string;
    known_duplicates_filename: string;  // will read existing duplicates from here
    new_duplicates_filename: string;    // will save newly found duplicates here
    scan_state: (all_files, master_scan, filtered_scan);
    skip_sysbckup_folder: boolean;  // true if we ignore Win 95 backup files
    skip_desktop_ini_folders: boolean;  // true if ignore special folders
    skip_this_folder: boolean;      // true if the current folder is to be skipped
    process_empty_files: boolean;   // true if not ingoring zero-length files
    check_file_timestamp: boolean;  // false for proper checksum comparison
    check_filename: boolean;        // true if file name is to be checked
    check_file_size: boolean;       // true if file size matters
    progress_bar: TProgressBar;     // component added at run-time
    min_width, min_height: integer;
    procedure find_files_to_compare;
    procedure compare_files;
    procedure update_compare_display;
    function do_checksum (size: integer;  const filename: String): TChecksum;
    function files_match (size: integer;  const name1, name2: String): boolean;
    function parse_list_view (lvw: TListView): string;
    procedure tidy_file_list_after_delete (first: TListItem);
    procedure update_caption;
  protected
    procedure GetMinMaxInfo (var info: TWMGetMinMaxInfo);  message WM_GETMINMAXINFO;
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;


implementation

uses ShellAPI, Registry, FilProps;

{$R *.DFM}

const
  saved_filename = 'FindDupl.lis';        // where the duplicate list is saved
  new_filename = 'NewDupl.lis';           // these are ones we found this time
  known_filename = 'SkipDupl.lis';        // these are ones we ignore on this run
  duplicate_group_prefix = 'Duplicate files .....';
  profile_path = '\Software\David J Taylor\FindDupl';
  folder_ignore_list_key = 'FoldersToIgnore';
  file_mask_key = 'FileMask';
  root_folder_key = 'RootFolder';
  floppy_is_master_key = 'FloppyIsMaster';
  skip_sysbckup_key = 'SkipSysbckupFolder';
  check_file_timestamp_key = 'CheckFileStamp';
  check_file_name_key = 'CheckFileName';
  check_file_size_key = 'CheckFileSize';
  skip_empty_files_key = 'SkipEmptyFiles';
  skip_desktop_folders_key = 'SkipDesktopFolders';


procedure TfrmMain.FormCreate(Sender: TObject);
begin
  // limit the minimum form size
  min_width := Width - 10;
  min_height := Height - 50;

  // Create the list to hold the file sizes and names found from the
  // folder scan.  This must be a sorted list to keep same sized or same named
  // files together - a fundamental requirement.  The "object" field
  // contains pointers to the DIR_LIST
  file_list := TStringList.Create;
  with file_list do
    begin
    Sorted := True;
    Duplicates := dupAccept;
    end;

  skip_file_list := TStringList.Create;
  with skip_file_list do
    begin
    Sorted := True;
    Duplicates := dupIgnore;
    end;

  skip_folder_list := TStringList.Create;
  with skip_folder_list do
    begin
    Sorted := True;
    Duplicates := dupIgnore;
    end;

  with TRegIniFile.Create (profile_path) do
    begin
    edtSkipFolder.Text := ReadString ('', folder_ignore_list_key, '');
    edtFileMask.Text := ReadString ('', file_mask_key, '*.*');
    edtRootDir.Text := ReadString ('', root_folder_key, '\');
    chkFloppyIsMaster.Checked := ReadBool ('', floppy_is_master_key, True);
    chkSkipSysbckup.Checked := ReadBool ('', skip_sysbckup_key, True);
    chkCheckFileTimestamp.Checked := ReadBool ('', check_file_timestamp_key, True);
    chkCheckFileName.Checked := ReadBool ('', check_file_name_key, False);
    chkCheckFileSize.Checked := ReadBool ('', check_file_size_key, True);
    chkSkipEmptyFiles.Checked := ReadBool ('', skip_empty_files_key, True);
    chkSkipDesktopFolders.Checked := ReadBool ('', skip_desktop_folders_key, True);
    Free;
    end;

  PageControl1.ActivePage := TabSheetBasic;

  // list of folders
  dir_list := TStringList.Create;
  dir_list.Sorted := False;

  // master list - contains just the master sizes
  master_list := TStringList.Create;
  with master_list do
    begin
    Sorted := True;
    Duplicates := dupAccept;
    end;

  progress_bar := TProgressBar.Create (Self);
  with progress_bar do
    begin
    Parent := StatusBar1;
    Left := 0;
    Top := 2;
    Width := StatusBar1.Panels [0].Width;
    Height := StatusBar1.Height - 2;
    Visible := False;
    end;

  saved_duplicates_filename :=
                   ExtractFileDir (Application.ExeName) + '\' + saved_filename;
  known_duplicates_filename :=
                   ExtractFileDir (Application.ExeName) + '\' + known_filename;
  new_duplicates_filename :=
                   ExtractFileDir (Application.ExeName) + '\' + new_filename;

  lblVersion.Caption := 'Find Duplicates version: V' +
                         VersionInfoResource1.FileVersion.AsString + ' ';
  comparing := False;
end;


procedure TfrmMain.FormDestroy(Sender: TObject);
var
  i: integer;
  sl: TStringList;
  file_names: TStringList;
begin
  // save the contents of the current search - if not blank
  // convert the ListView into a string list
  if FileListView1.Items.Count > 0 then     // do the save
    begin
    sl := TStringList.Create;
    file_names := TStringList.Create;
    with FileListView1.Items do
      for i := 0 to Count - 1 do       // scan down the list view
        with Item [i] do               // the current item (row)
          begin
          if SubItems.Count = 0        // are there any sub-items?
          then sl.Add ('"' + Caption + '"')    // no - just save the caption
          else sl.Add ('"' + Caption + '",' + SubItems.CommaText);  // yes - save the lot
          if (Length (Caption) > 2) and (Caption [2] = ':')
            then file_names.Add (Caption);
          end;
    // just in case the "save" fails (e.g. read-only folder or file)
    try
      sl.SaveToFile (saved_duplicates_filename);
      file_names.SaveToFile (new_duplicates_filename);
    except
    end;
    sl.Free;
    file_names.Free;
    end;

  // return the items we allocated in the Create method
  progress_bar.Free;
  master_list.Free;
  skip_file_list.Free;
  skip_folder_list.Free;
  file_list.Free;
  dir_list.Free;
end;


procedure TfrmMain.GetMinMaxInfo (var info: TWMGetMinMaxInfo);
begin
  with info.MinMaxInfo.ptMinTrackSize do
    begin
    x := min_width;
    y := min_height;
    end;
end;


procedure TfrmMain.FormResize(Sender: TObject);
begin
  // Handle all the resizing stuff.
  // We could do this with panels rather than all this computation.....
  // position the search controls group box
  PageControl1.Left := frmMain.ClientWidth - 4 - PageControl1.Width;

  // size the what to search for group box
  GroupBox1.Width := PageControl1.Left - 20;

  // position the buttons in this group box
  btnExit.Left := GroupBox1.Width - 8 - btnExit.Width;
  btnSearch.Left := btnExit.Left - 16 - btnSearch.Width;

  // position the root folder and root file edit boxes, and their captions
  edtRootDir.Left := GroupBox1.Width - 8 - edtRootDir.Width;
  edtFileMask.Left := edtRootDir.Left;
  edtSkipFolder.Left := edtRootDir.Left;
  Label7.Left := edtRootDir.Left - Label7.Width;
  Label2.Left := edtFileMask.Left - Label2.Width;
  Label1.Left := edtSkipFolder.Left - Label1.Width;

  // expand the disk list view to fill the remaining available width
  DiskListView1.Width := btnSearch.Left - 24;

  // position the version label
  lblVersion.Left := frmMain.ClientWidth - lblVersion.Width - 4;
  lblVersion.Top := frmMain.ClientHeight - StatusBar1.Height - lblVersion.Height;

  // fill the lower half of the form with the list of duplicates found
  FileListView1.Width := frmMain.ClientWidth - 12;
  FileListView1.Top := lblFileListCaption.Top + lblFileListCaption.Height + 2;
  FileListView1.Height := frmMain.ClientHeight - FileListView1.Top -
                             lblVersion.Height - StatusBar1.Height - 2;
end;


procedure TfrmMain.Timer1Timer(Sender: TObject);
var
  i: integer;
  j: integer;
  s: TSearchRec;
  sl_items: TStringList;
  sl_sub_items: TStringList;
  new_item: TListItem;
  in_group: boolean;
  item_caption: string;
  column: integer;
begin
  // This routine contains all the initialisation code that is handled in the
  // context of the running program.  First we check for the results of a
  // previous run, then fill in the Drives list view.
  Timer1.Enabled := False;

  // Is there a list of known duplicates?
  if FindFirst (known_duplicates_filename, faAnyFile, s) = 0 then
    begin
    // Does the user want to load the data from the previous run?
    if Application.MessageBox (PChar (
          'A list of known duplicate files to skip was found from a'#13 +
          FormatDateTime ('"previous run on" yyyy-mmm-dd "at" hh:nn:ss',
                          FileDateToDateTime (s.Time)) + #13#13 +
          'Reloading that list will prevent these known ' +
          'duplicates from being found again.' + #13#13 +
          'Do you want to reload that list?'),
          'Find Duplicate Files - known duplicates to skip list found',
          MB_YESNO or MB_ICONQUESTION) = IDYES then
      try
        skip_file_list.LoadFromFile (known_duplicates_filename);
      except
      end;
    end;
  FindClose (s);

  // Was there a previous run?
  if FindFirst (saved_duplicates_filename, faAnyFile, s) = 0 then
    begin
    // Does the user want to load the data from the previous run?
    if Application.MessageBox (PChar (
          'A list of duplicate files was found from a'#13 +
          FormatDateTime ('"previous run on" yyyy-mmm-dd "at" hh:nn:ss',
                          FileDateToDateTime (s.Time)) + #13#13 +
          'Reloading that list will save time.' + #13#13 +
          'Do you want to reload that list?'),
          'Find Duplicate Files - previous list found',
          MB_YESNO or MB_ICONQUESTION) = IDYES then
      begin
      sl_items := TStringList.Create;    // list containing an image of the file
      StatusBar1.Update;
      sl_items.LoadFromFile (saved_duplicates_filename);   // load the list
      progress_bar.Visible := True;
      progress_bar.Min := 0;
      progress_bar.Max := sl_items.Count - 1;
      Screen.Cursor := crHourGlass;
      FileListView1.Items.BeginUpdate;
      for column := 0 to FileListView1.Columns.Count - 1 do
        FileListView1.Columns [column].Width := 0;
      in_group := False;
      sl_sub_items := TStringList.Create;   // temp list for each line
      for i := 0 to sl_items.Count - 1 do   // scan down the saved list view
        begin
        progress_bar.Position := i;
        new_item := FileListView1.Items.Add;     // add the next item in the list view
        // assume each string in the file contains comma text for each item(list)
        sl_sub_items.CommaText := sl_items.Strings [i];
        // assume we'll at least have a caption, even though it could be blank
        item_caption := Trim (sl_sub_items.Strings [0]);
        // reset IN_GROUP if we've just read a caption
        if Length (item_caption) = 0 then in_group := False;
        new_item.Caption := item_caption;
        // scan across the remaining items in the list, adding as sub-items

        for j := 1 to sl_sub_items.Count - 1 do
          new_item.SubItems.Add (Trim (sl_sub_items [j]));

        // in a list of duplicates? - then bump the file an byte count
        // just in case the string isn't a valid number, use try/except block
        if in_group then
          try
            Inc (duplicate_files);
            Inc (duplicate_bytes, StrToInt (sl_sub_items [1]));
          except
          end;

        // only now do we see if we might be in a group of duplicates
        in_group := sl_sub_items.Count > 1;
        end;

      for column := 0 to FileListView1.Columns.Count - 1 do
        FileListView1.Columns [column].Width := ColumnTextWidth;
      // done reading the file, clean up
      sl_items.Free;
      sl_sub_items.Free;
      FileListView1.Items.EndUpdate;
      Screen.Cursor := crDefault;
      progress_bar.Visible := False;
      StatusBar1.Panels [0].Text := 'Duplicates read from file .....';
      update_compare_display;
      end;
    end;
  FindClose (s);

  // now find out what drives are available, add an icon and the display name to
  // the Drives list view.  If a drive is local, pre-select it for the user.
  DiskListView1.ScanDrives;

//  ShowMessage (VersionInfoResource1.FileVersion);
end;


procedure TfrmMain.DiskListView1Change(Sender: TObject; Item: TListItem;
  Change: TItemChange);
// this procedure updates the main form's caption to reflect the drives selected
begin
  update_caption;
end;


procedure TfrmMain.edtRootDirExit(Sender: TObject);
begin
  update_caption;
end;


procedure TfrmMain.update_caption;
// this procedure updates the main form's caption to reflect
// the drives and folders selected
var
  s, t: string;
  i: integer;
begin
  s := DiskListView1.SelectedDrives;
  t := 'Find Duplicate Files named ' + edtFileMask.Text + ' on ';
  for i := 1 to Length (s) do t := t + s [i] + ': ';
  t := t + 'in ' + edtRootDir.Text;
  Caption := t;
end;


procedure TfrmMain.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
// just note that we are required to close
begin
  TreeScanner1.Continue := False;
  continue_compare := False;
  with TRegIniFile.Create (profile_path) do
    begin
    WriteString ('', folder_ignore_list_key, edtSkipFolder.Text);
    WriteString ('', file_mask_key, edtFileMask.Text);
    WriteString ('', root_folder_key, edtRootDir.Text);
    WriteBool ('', floppy_is_master_key, chkFloppyIsMaster.Checked);
    WriteBool ('', skip_sysbckup_key, chkSkipSysbckup.Checked);
    WriteBool ('', check_file_timestamp_key, chkCheckFileTimestamp.Checked);
    WriteBool ('', check_file_name_key, chkCheckFileName.Checked);
    WriteBool ('', check_file_size_key, chkCheckFileSize.Checked);
    WriteBool ('', skip_empty_files_key, chkSkipEmptyFiles.Checked);
    WriteBool ('', skip_desktop_folders_key, chkSkipDesktopFolders.Checked);
    Free;
    end;
end;


procedure TfrmMain.btnCloseClick(Sender: TObject);
begin
  Close;
end;


procedure TfrmMain.btnSearchClick(Sender: TObject);
begin
  if TreeScanner1.SearchInProgress
  then                                 // must be in folder scan phase
    begin
    TreeScanner1.Continue := False;    // stop the tree scanner
    continue_compare := False;         // and stop looping over drives
    end
  else                                 // in comparison phase
    if comparing
    then
      continue_compare := False        // stop the comparison
    else
      begin
      lblFileListCaption.Caption := 'Searching ...';
      btnSearch.Caption := 'Stop';     // prepare for a stop
      continue_compare := True;
      find_files_to_compare;           // build list of files sorted by size or name
      compare_files;                   // and compare the contents of those files
      end;
end;


procedure TfrmMain.find_files_to_compare;
// Find all the files in the tree that match the specification on all the
// selected drives.  The files are sorted by size.

var
  i: integer;
  root_folder_list: TStringList;
  drives: string;
  this_drive: string;
  dir: string;
  folder: integer;
  temp_list: TStringList;
  s1, s2: string;
begin
  with StatusBar1 do
    begin
    Panels [1].Text := 'Bytes:';
    Panels [2].Text := 'Files:';
    end;

  FileListView1.Hint := '';
  // clear out any existing matches
  Screen.Cursor := crHourGlass;
  FileListView1.Items.Clear;
  Screen.Cursor := crDefault;
  file_list.Clear;
  dir_list.Clear;
  master_list.Clear;
  root_folder_list := TStringList.Create;
  with root_folder_list do
    begin
    Sorted := True;
    Duplicates := dupIgnore;
    end;

  temp_list := TStringList.Create;
  with temp_list do
    begin
    Sorted := True;
    Duplicates := dupIgnore;
    end;

  // Ensure that all items on the skip folder list end with "\"
  // and save the list back into the EditBox
  if Trim (edtRootDir.Text) = '' then edtRootDir.Text := '\';
  root_folder_list.CommaText := edtRootDir.Text;
  for i := 0 to root_folder_list.Count - 1 do
    begin
    dir := Trim (LowerCase (root_folder_list.Strings [i]));
    if dir [1] <> '\' then dir := '\' + dir;
    if dir [Length (dir)] <> '\' then dir := dir + '\';
    temp_list.Add (dir);
    end;

  // We need to ensure that each folder in the list is not a sub-folder of one
  // earlier in the list.  Start with the first entry.  Use a simple method as we
  // don't expect a great many folders to be entered by the user.
  // Use upper case display to make the display more prominent.
  folder := 0;                          // index of first folder in the list
  while folder < temp_list.Count-1 do   // must be at least one item more to check
    begin
    s1 := temp_list.Strings [folder];  // earlier candidate folder
    i := folder + 1;                    // index of first string to check
    while i <= temp_list.Count-1 do     // still items to check remaining
      begin
      s2 := temp_list.Strings [i];     // current folder name
      if Pos (s1, s2) <> 1              // is the earlier folder in the current?
      then Inc (i)                      // no - go and check the next folder
      else                              // yes - remove the contained sub-folder
        begin                           // tell the user why we won't scan
        ShowMessage ('Folder: ' + UpperCase (s2) + '  is already included in ' +
          #10 + 'folder: ' + UpperCase (s1) + '  and will therefore be removed.');
        temp_list.Delete (i);           // remove the contained sub-folder
        end;
      end;
    Inc (folder);                       // point to the next folder to check
    end;

  edtRootDir.Text := temp_list.CommaText;
  update_caption;
  root_folder_list.CommaText := edtRootDir.Text;

  // Ensure that all items on the skip folder list begin and end with "\"
  // and save the list back into the EditBox
  skip_folder_list.CommaText := edtSkipFolder.Text;
  temp_list.Clear;
  for i := 0 to skip_folder_list.Count - 1 do
    begin
    dir := Trim (LowerCase (skip_folder_list.Strings [i]));
    if dir [1] <> '\' then dir := '\' + dir;
    if dir [Length (dir)] <> '\' then dir := dir + '\';
    // make sure a folder really has been specified.....
    if Length (dir) >= 3 then temp_list.Add (dir);
    end;
  edtSkipFolder.Text := temp_list.CommaText;
  skip_folder_list.CommaText := edtSkipFolder.Text;
  temp_list.Free;

  if edtFileMask.Text = '' then edtFileMask.Text := '*.*';
  // set as if for multiple files
  TreeScanner1.FileMaskCommaText := edtFileMask.Text;
  // read-back list as it should be formatted as visual clue for user
  edtFileMask.Text := TreeScanner1.FileMaskCommaText;
  skip_sysbckup_folder := chkSkipSysbckup.Checked;
  skip_desktop_ini_folders := chkSkipDesktopFolders.Checked;

  process_empty_files := not chkSkipEmptyFiles.Checked;
  check_file_size := chkCheckFileSize.Checked;

  drives := DiskListView1.SelectedDrives;   // get string of all selected drives
  for i := 1 to Length (drives) do          // do one drive letter at a time
    begin
    this_drive := drives [i] + ':';         // build a valid drive string
    scan_state := all_files;                // assume normal scan to start with
    // if we've just built the master list, all subsequent scans are filtered
    if scan_state = master_scan then scan_state := filtered_scan;
    // if we're just starting and this is a floppy, get the master list
    if (scan_state = all_files) and
       (chkFloppyIsMaster.Checked) and
       (this_drive = 'A:') then scan_state := master_scan;
    with StatusBar1.Panels [0] do           // report scan state to user
      case scan_state of
            all_files: Text := 'Scanning folders ....';
          master_scan: Text := 'Scanning master folders ....';
        filtered_scan: Text := 'Scanning filtered folders ....';
      end;
    // form the root scan path by drive + folder
    for folder := 0 to root_folder_list.Count-1 do
      begin
      TreeScanner1.InitialDirectory := this_drive + root_folder_list.Strings [folder];
      TreeScanner1.ScanTree;         // scan a folder tree on the selected drive
      end;
    if not continue_compare then Break;     // check if we can continue
    end;

  root_folder_list.Free;
  // clear out any messages from scanning
  StatusBar1.Panels [0].Text := '';
  StatusBar1.Panels [3].Text := '';
end;


procedure TfrmMain.TreeScanner1DirectoryFound(Sender: TObject);
// this procedure is called back by the tree scanner for each folder
var
  dir_name: string;
  i: integer;
begin
  dir_name := Trim (TreeScanner1.FileFound);  // get the folder name
  dir_list.Add (dir_name);                    // and add it to the list of dirs
  current_dir := dir_list.Count - 1;          // taking note of the index
  // determine if we should skip this folder (because its' Win 95 System Backup)
  skip_this_folder := (skip_sysbckup_folder) and
                      (Pos ('\SYSBCKUP\', UpperCase (dir_name)) <> 0);
  // check if any of the specified folder names to skip match this folder
  for i := 0 to skip_folder_list.Count - 1 do
    skip_this_folder := skip_this_folder or
                      (Pos (skip_folder_list.Strings [i], LowerCase (dir_name)) <> 0);
  if skip_desktop_ini_folders then
    skip_this_folder := skip_this_folder or FileExists (dir_name + '\desktop.ini');
  if skip_this_folder
    then StatusBar1.Panels [3].Text := 'Skipping ' + dir_name
    else StatusBar1.Panels [3].Text := dir_name;
  file_list.BeginUpdate;                      // only really affects visible comps
end;


procedure TfrmMain.TreeScanner1DirectoryDone(Sender: TObject);
begin
  file_list.EndUpdate;
end;


procedure TfrmMain.TreeScanner1FileFound(Sender: TObject);
// this procedure is called back by the tree scanner for each file
var
  file_name: String;
  file_size: String;
  f: TSearchRec;
  dummy: integer;
begin
  if skip_this_folder then Exit;           // ignore skipped folders
  file_name := TreeScanner1.FileFound;     // save the file name
  if skip_file_list.Find (file_name, dummy) then Exit;
  // skip those pesky special files
  if Pos (LowerCase (ExtractFileName (file_name)), 'mm256.dat mm2048.dat') <> 0
    then Exit;
  // if we can find the file, add an entry to the file list with the size as a
  // numeric prefix (to allow sorting) and with the folder as an index into
  // the folder list (store the index as an integer, type-cast to a pointer)
  if FindFirst (file_name, faAnyFile, f) = 0 then
    begin
    if (process_empty_files) or (f.size <> 0) then
      begin
      if check_file_size
      then file_size := Format ('%9d ', [f.size])   // store the size
      else file_size := f.Name + '\';               // store the name
      if scan_state = master_scan then master_list.Add (file_size);
      // should we add this file?  Yes if doing all files, yes if scanning the
      // master folder, and yes if filtered and we get a match
      if (scan_state = all_files) or
         (scan_state = master_scan) or
         ((scan_state = filtered_scan) and (master_list.Find (file_size, dummy))) then
        begin
        file_size := file_size + ExtractFileName (file_name);    // add the name string
        file_list.AddObject (file_size, Pointer (current_dir));  // integer type-cast as pointer
        end;
      end;
    end;
  FindClose (f);
end;


procedure TfrmMain.update_compare_display;
// formatted update of duplicate bytes and files found
begin
  StatusBar1.Panels[1].Text := Format ('Bytes: %.0n', [duplicate_bytes + 0.0]);
  StatusBar1.Panels[2].Text := Format ('Files: %.0n', [duplicate_files + 0.0]);
end;


function TfrmMain.files_match (size: integer;  const name1, name2: String): boolean;
// primary file comparison routine, returns True if files match
// zero length files automatically match!
// updates number of files and bytes that matched
// uses memory-mapped file I/O
// could be improved by matching 32-bit quantities at a time, rather
// than the bytes used in this version
type
  PByte = ^byte;
var
  same: boolean;                   // true if files match
  handle1, handle2: THandle;       // stuff for memory-mapped file I/O
  mapping1, mapping2: THandle;
  base1, base2: Pointer;
  p1, p2: PByte;                   // pointers within each file
  i: integer;                      // loop over bytes in the file
begin
  if not check_file_size then
    begin
    Inc (duplicate_files);
    Result := True;
    Exit;
    end;

  if check_filename and
    (LowerCase (ExtractFileName (name1)) <> LowerCase (ExtractFileName (name2))) then
      begin
      Result := False;
      Exit;
      end;

  if size = 0 then                 // special case - zero length files match!
    begin
    Inc (duplicate_files);
    Result := True;
    Exit;
    end;

  Result := False;
  handle1 := CreateFile (PChar(name1), GENERIC_READ, FILE_SHARE_READ, nil,
                         OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
  if handle1 <> INVALID_HANDLE_VALUE then
    begin
    handle2 := CreateFile (PChar(name2), GENERIC_READ, FILE_SHARE_READ, nil,
                           OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if handle2 <> INVALID_HANDLE_VALUE then
      begin
      mapping1 := CreateFileMapping (handle1, nil, PAGE_READONLY, 0, 0, nil);
      if mapping1 <> 0 then
        begin
        mapping2 := CreateFileMapping (handle2, nil, PAGE_READONLY, 0, 0, nil);
        if mapping2 <> 0 then
          begin
          base1 := MapViewOfFile (mapping1, FILE_MAP_READ, 0, 0, 0);
          if base1 <> nil then
            begin
            base2 := MapViewOfFile (mapping2, FILE_MAP_READ, 0, 0, 0);
            if base2 <> nil then
              begin
              same := True;
              p1 := PByte (base1);         // point to start of first file
              p2 := PByte (base2);         // and second file
              i := 0;                      // initialise loop counter
              while same and (i < size) do
                begin
                same := p1^ = p2^;         // compare a single byte
                Inc (i);                   // bump the count
                Inc (p1);                  // and the two read pointers
                Inc (p2)
                end;
              Result := same;              // record the outcome
              if same then
                begin
                Inc (duplicate_bytes, size);
                Inc (duplicate_files);
                end;
              UnmapViewOfFile (base2);     // undo all the mapped I/O stuff
              end;
            UnmapViewOfFile (base1);
            end;
          CloseHandle (mapping2);
          end;
        CloseHandle (mapping1);
        end;
      CloseHandle (handle2);
      end;
    CloseHandle (handle1);
    end;
end;


function TfrmMain.do_checksum (size: integer;  const filename: String): TChecksum;
// checks either date/time or the first 512 bytes of the file
// uses memory-mapped I/O
// zero byte files return a zero checksum
const
  bytes_per_sector = 512;
  max_dwords_in_checksum = bytes_per_sector div SizeOf (TChecksum);
var
  checksum: TChecksum;
  handle: THandle;
  mapping: THandle;
  base: Pointer;
  dwords_to_checksum: integer;
  p: ^DWORD;
  dw: DWORD;
  search: TSearchRec;
begin
  Result := 0;
  if (size = 0) or (not check_file_size) then Exit;    // special cases

  checksum := 0;

  if check_file_timestamp   // see what method is to be used
  then
    begin
    if FindFirst (filename, faAnyFile, search) = 0 then checksum := search.Time;
    FindClose (search);
    end
  else
    begin
    handle := CreateFile (PChar(filename), GENERIC_READ, FILE_SHARE_READ, nil,
                          OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if handle <> INVALID_HANDLE_VALUE then
      begin
      mapping := CreateFileMapping (handle, nil, PAGE_READONLY, 0, 0, nil);
      if mapping <> 0 then
        begin
        base := MapViewOfFile (mapping, FILE_MAP_READ, 0, 0, 0);
        if base <> nil then
          begin
          p := base;
          dwords_to_checksum := size div SizeOf (DWORD);
          if dwords_to_checksum > max_dwords_in_checksum then
            dwords_to_checksum := max_dwords_in_checksum;
          for dw := 1 to dwords_to_checksum do
            begin
            checksum := checksum + p^;
            Inc (p);
            end;
          UnmapViewOfFile (base);
          end;
        CloseHandle (mapping);
        end;
      CloseHandle (handle);
      end;
    end;
  // There's a chance that the total has hit -2147483648, which is not a valid
  // integer when you use StrToInt.  To avoid this, kill the top bit and return
  // a 31-bit result.  It doesn't matter that the value isn't a true sum,
  // it _will_ be the same for each summation which is what matters.
  Result := checksum and $7FFFFFFF;
end;


procedure TfrmMain.compare_files;
// This is the main procedure to compare all files found from the folder
// scan.  This is sorted by size, so we now look for blocks of files that
// have the same size.  For each of these blocks of multiples, create a list
// with the file initial checksum and the file name and pass that to an
// inner procedure.
//
// Measurements show that the critical routine here (in terms of execution time)
// is the Checksum function, hence its default replacement by a simple
// date and time check.
var
  items_done: integer;             // count for progress bar display
  columns_frozen: boolean;

  procedure freeze_column_widths;
  var
    col: integer;
  begin
    // freeze the list view columns at their current value
    for col := 0 to FileListView1.Columns.Count - 1 do
      with FileListView1.Columns [col] do
        begin
        Width := ColumnHeaderWidth;
        Width := Width;
        end;
    FileListView1.Refresh;
    columns_frozen := True;
  end;

  procedure unfreeze_column_widths;
  var
    col: integer;
  begin
    // freeze the list view columns at their current value
    for col := 0 to FileListView1.Columns.Count - 1 do
      with FileListView1.Columns [col] do Width := ColumnHeaderWidth;
    columns_frozen := False;
  end;

  procedure do_multiples (size: integer;  list: TStringList);
  // handle a list of files of the same size
  var
    master_filename: string;
    old_filename: string;
    checksum: string;
    old_checksum: string;
    in_checksum_group: boolean;
    in_match_group: boolean;
    i: integer;
    s: String;
    space: integer;
    same_checksum: boolean;
    exact_match: boolean;
    new_item: TListItem;
    filesize: integer;
  begin
    old_filename := '';                  // set to impossible initial values
    old_checksum := '';                  // unlikely, if not impossible!
    in_checksum_group := False;          // initially, not in checksum group
    in_match_group := False;             // nor in match group

    if not check_file_size then progress_bar.Max := list.Count - 1;

    // scan down the list of files which are the same size
    for i := 0 to list.Count - 1 do
      begin
      s := Trim (list.Strings [i]);       // separate out the size part of the string

      // split into components
      if check_file_size
      then
        begin                                     //  "12345 \full-path"
        space := Pos (' ', s);                    // locate the space
        checksum := Copy (s, 1, space-1);         // get the checksum part
        end
      else
        begin                                     // "filename.dat\ full-path"
        space := Pos ('\', s) + 1;                // locate the end of the file name
        checksum := UpperCase (Copy (s, 1, space-2));  // get the checksum part
        end;

      s := Trim (Copy (s, space, 999));           // and the file name part
      same_checksum := checksum = old_checksum;   // is this the same as before?

      if same_checksum and (not in_checksum_group)
        then master_filename := old_filename;

      old_checksum := checksum;
      old_filename := s;

      if same_checksum
        then exact_match := files_match (size, master_filename, s)
        else exact_match := False;

      in_checksum_group := same_checksum;

      if (not in_match_group) and exact_match then
        begin
        if not check_file_size then
          begin
          progress_bar.Position := i;
          update_compare_display;
          StatusBar1.Panels [3].Text := 'Comparing files named: ' + checksum;
          end;
        with FileListView1 do
          begin
          if Items.Count <> 0 then Items.Add;  // add a blank
          new_item := Items.Add;               // add a title line for the group
          new_item.Caption := duplicate_group_prefix;
          AddFile (master_filename, filesize);
          end;
        end;

      in_match_group := exact_match;

      if in_checksum_group and exact_match
        then FileListView1.AddFile (s, filesize);         // and note this duplicate
        
      Application.ProcessMessages;
      if not continue_compare then Break;
      end;

    update_compare_display;                          // the bytes and files
    progress_bar.Position := items_done;             // show the progress
    if (not columns_frozen) and (duplicate_files > 0) then freeze_column_widths;
    SendMessage (FileListView1.Items.Handle, WM_SETREDRAW, 0, 0);
    SendMessage (FileListView1.Items.Handle, WM_SETREDRAW, 1, 0);
  end;

const
  fixed_file_size = 1234567890;   // nominal file size when only name matters
var
  current_size: integer;
  old_size: integer;
  old_filename: String;
  current_filename: String;
  item: integer;
  s: String;
  in_multiple: boolean;
  same_size: boolean;
  multiples_list: TStringList;
  chk: TChecksum;
  space: integer;
begin
  // scan the list of files finding all the files of duplicate size
  comparing := True;
  duplicate_bytes := 0;     // initially, no duplicate files
  duplicate_files := 0;     // or bytes duplicated
  old_size := -1;
  old_filename := '';
  in_multiple := False;

  unfreeze_column_widths;
  FileListView1.Items.BeginUpdate;     // stop the visible updates

  // each group of similarly sized files will be kept in this sorted list
  // except that we keep files in as-supplied order when ignoring file size
  multiples_list := TStringList.Create;
  with multiples_list do
    begin
    if check_file_size then Sorted := True;
    Duplicates := dupAccept;
    end;

  if file_list.Count = 0
  then progress_bar.Max := 1
  else progress_bar.Max := file_list.Count - 1;  // the maximum number of files to check
  progress_bar.Visible := True;             // show the progress bar
  items_done := 0;                          // with nothing done, as yet

  check_file_timestamp := chkCheckFileTimestamp.Checked;  // note the checksum method
  check_filename := chkCheckFilename.Checked;

  lblFileListCaption.Caption := 'Comparing files ...';
  FileListView1.Items.Clear;

  // Use a WHILE loop to scan the list rather than a DO to allow for the
  // comparison to be interrupted.  Start with the last item which has the
  // largest file size, so that we record the worst-case duplicates first.
  item := file_list.Count - 1;
  while (item >= 0) and continue_compare do
    begin
    s := Trim (file_list.Strings [item]);    // get the current size and filename
    // split into components
    if check_file_size
    then space := Pos (' ', s)
    else space := Pos ('\', s) + 1;
    // retrieve the folder name from the stored list, the index into the
    // list is given by the integer, stored as a pointer in the Object field
    current_filename := dir_list.Strings [Integer (file_list.Objects [item])] +
                        Trim (Copy (s, space, 999));  // add the file name
    Delete (s, space, 999);                  // retain just the size/name string
    if check_file_size
    then current_size := StrToInt (s)        // convert the file size to integer
    else current_size := fixed_file_size;    // use a nominal figure

    same_size := current_size = old_size;    // is size the same as before ?
    if same_size and (not in_multiple) then  // start of a new set of files
      begin
      multiples_list.Clear;                  // clear out any existing stuff
      if check_file_size
      then
        begin
        chk := do_checksum (old_size, old_filename);   // get the checksum and add to list
        multiples_list.Add (Format ('%12d ', [chk]) + ' ' + old_filename);
        StatusBar1.Panels [3].Text := 'Comparing files with the same name ...';
        end
      else
        begin
        multiples_list.Add (ExtractFileName (old_filename) + '\ ' + old_filename);
        StatusBar1.Panels [3].Text := 'Comparing ' +
          Format ('%.0n', [current_size + 0.0]) + ' byte files ...';
        end;
      // update the user on the progress of the comparison
      StatusBar1.Update;
      end;

    in_multiple := same_size;
    if in_multiple
    then
      // if in a list, checksum this file and add it to the list
      begin
      if check_file_size
      then
        begin
        chk := do_checksum (current_size, current_filename);
        multiples_list.Add (Format ('%12d ', [chk]) + ' ' + current_filename);
        end
      else
        multiples_list.Add (ExtractFileName (current_filename) + '\ ' + current_filename);
      end
    else
      // if not in a list, process any existing list and reset it by clearing out
      begin
      if multiples_list.Count <> 0 then
        begin
        do_multiples (old_size, multiples_list);
        multiples_list.Clear;
        end;
      end;

    old_size := current_size;
    old_filename := current_filename;
    Inc (items_done);              // update the progress
    file_list.Delete (item);       // give the memory back for this item
    Dec (item);
    Application.ProcessMessages;   // look for any button press
    end;

  // we might get here and still have a list of multiples pending
  // so process it, but there's no point in clearing out the list
  if multiples_list.Count <> 0 then
    do_multiples (old_size, multiples_list);

  lblFileListCaption.Caption :=
              'What duplicates were found - double-click a file to review ...';
  FileListView1.Hint := 'Double-click a file to review';

  // try and make the status text reflect the most recent user command
  if TreeScanner1.Continue
    then StatusBar1.Panels [3].Text := 'Scan complete'
    else StatusBar1.Panels [3].Text := 'Scan interrupted';

  if not continue_compare
    then StatusBar1.Panels [3].Text := 'Comparison interrupted';

  if duplicate_files <> 0
    then StatusBar1.Panels [0].Text := 'Duplicates found....';

  // clear up after this procedure
  Screen.Cursor := crHourGlass;        // this could take some time .....
  FileListView1.Items.EndUpdate;       // stop the visible updates
  unfreeze_column_widths;              // allow auto-sizing for the file list view
  Screen.Cursor := crDefault;
  progress_bar.Visible := False;       // hide the progress bar display
  progress_bar.Position := 0;          // reset for next time
  multiples_list.Free;                 // give back the memory
  file_list.Clear;
  dir_list.Clear;
  comparing := False;
  btnSearch.Caption := 'Find Now';
end;


function TfrmMain.parse_list_view (lvw: TListView): string;
// function to return filename from selected TListView item caption
// ignores those items starting "Duplicate group..."
var
  filename: string;
begin
  Result := '';
  with lvw do
    begin
    if Selected = nil then Exit;            // nothing selected - skip
    filename := Trim (Selected.Caption);    // get the caption string
    if Length (filename) = 0 then Exit;     // blank - skip
    if Pos (duplicate_group_prefix, filename) <> 0 then Exit;
    Result := filename;                     // return the file name
    end;
end;


procedure TfrmMain.tidy_file_list_after_delete (first: TListItem);
var
  middle, last: TListItem;
  num_files: integer;
begin
  // We may have changed the list, so check for the number in a group
  // scan up the list until we find the "start of group" message
  while Pos (duplicate_group_prefix, first.Caption) = 0 do
    first := FileListView1.GetNextItem (first, sdAbove, [isNone]);
  last := first;
  middle := nil;
  num_files := 0;

  // scan down the list until a blank is found or we fall off the list end
  // count the number of files remaining
  repeat
    last := FileListView1.GetNextItem (last, sdBelow, [isNone]);
    if last <> nil then              // not off the list end
      if last.Caption <> '' then     // still in group of files
        begin
        Inc (num_files);
        middle := last;              // save pointer to real file name entry
        end;
  until (last = nil) or (last.Caption = '');

  // only one file left? - then we delete the header, the file and any trailer
  if num_files = 1 then
    with FileListView1 do
      begin
      if last <> nil then Items.Delete (last.index);
      Items.Delete (middle.index);
      Items.Delete (first.index);
      end;

  FileListView1.Hint := 'Double-click a file to review';
end;


procedure TfrmMain.FileListView1DblClick(Sender: TObject);
// Handle a double-click, get the filename and display properties to the user
var
  filename: string;
  first: TListItem;
begin
  filename := parse_list_view (FileListView1);
  if filename = '' then Exit;
  first := FileListView1.GetNextItem (FileListView1.Selected, sdAbove, [isNone]);

  if FileExists (filename)
  then
    begin
    frmFileProperties1.filename := filename;
    if frmFileProperties1.ShowModal = idYes then FileListView1.Selected.Delete;
    end
  else
    FileListView1.Selected.Delete;

  tidy_file_list_after_delete (first);
end;


procedure TfrmMain.FileListView1Click(Sender: TObject);
// handle a single-click, make the hint echo the file name or reset to default
var
  filename: string;
begin
  filename := parse_list_view (FileListView1);
  with FileListView1 do
    if filename = ''
    then Hint := 'Double-click a file to review'
    else Hint := filename;
end;


procedure TfrmMain.btnExitClick(Sender: TObject);
begin
  Close;
end;


procedure TfrmMain.FileListView1ColumnClick(Sender: TObject;
  Column: TListColumn);
begin
// override event - we don't want the user sorting the data
end;


procedure TfrmMain.chkCheckFilesizeClick(Sender: TObject);
begin
  if not chkCheckFilesize.Checked
  then
    begin
    chkCheckFilename.Checked := True;
    chkCheckFileTimestamp.Checked := False;
    chkCheckFilename.Enabled := False;
    chkCheckFileTimestamp.Enabled := False;
    end
  else
    begin
    chkCheckFilename.Enabled := True;
    chkCheckFileTimestamp.Enabled := True;
    end;
end;


// Based on code from Chris van den Hout for handling a Delete key press

procedure TfrmMain.FileListView1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);

  function DeleteOrTrashFile (filename: string;  ShiftDown: boolean): boolean;
  // Procedure to remove a file.
  // If Shift is held down, really Delete the file rather than sending
  // it to the recycling bin.  This mimicks Windows behaviour
  var
    shell_filename: string;
    SHFileOp: TSHFileOpStruct;
  begin
    Result := False;
    shell_filename := filename + #0#0;
    FillChar (SHFileOp, SizeOf (SHFileOp), 0);
    with SHFileOp do
      begin
      wFunc := FO_DELETE;
      pFrom := PChar (shell_filename);
      if not ShiftDown then fFlags := FOF_ALLOWUNDO;
      end;                                          
    SHFileOperation (SHFileOp);
    if not FileExists (filename) then Result := True;
  end;

var
  filename: string;
  first: TListItem;
begin
  if Key = VK_DELETE then
    begin
    filename := parse_list_view (FileListView1);
    if filename = '' then Exit;
    first := FileListView1.GetNextItem (FileListView1.Selected, sdAbove, [isNone]);

    if DeleteOrTrashFile (filename, (Shift = [ssShift])) then
      FileListView1.Selected.Delete;

    tidy_file_list_after_delete (first);
    end;
end;

end.

